"""
Testing the properties of downscaling methods
=============================================

1. Tests that means are preserved by quantile mapping
2. Test that trends are preserved by quantile mapping
3. ... ?

"""
# TODO
